# flake8: noqa
from .auth_views import login_for_access_token
from .extra_views import docs, favicon
from .feature_views import get_features_handler
from .tags_views import tags_item_handler, tags_list_handler
from .testuser_views import test_user_get_spec_handler, test_user_handler, test_user_put_spec_handler



import abc
from typing import cast

from overhave import db
from overhave.storage.converters import FeatureTypeModel


class BaseFeatureTypeStorageException(Exception):
    """Base exception for :class:`FeatureTypeStorage`."""


class FeatureTypeNotExistsError(BaseFeatureTypeStorageException):
    """Exception for situation without specified feature type."""


class IFeatureTypeStorage(abc.ABC):
    """Abstract class for feature type storage."""

    @staticmethod
    @abc.abstractmethod
    def get_default_feature_type() -> FeatureTypeModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_feature_type_by_name(name: str) -> FeatureTypeModel:
        pass


class FeatureTypeStorage(IFeatureTypeStorage):
    """Class for feature type storage."""

    @staticmethod
    def get_default_feature_type() -> FeatureTypeModel:
        with db.create_session() as session:
            feature_type: db.FeatureType = session.query(db.FeatureType).order_by(db.FeatureType.id.asc()).first()
            return cast(FeatureTypeModel, FeatureTypeModel.from_orm(feature_type))

    @staticmethod
    def get_feature_type_by_name(name: str) -> FeatureTypeModel:
        with db.create_session() as session:
            feature_type: db.FeatureType = (
                session.query(db.FeatureType).filter(db.FeatureType.name == name).one_or_none()
            )
            if feature_type is None:
                raise FeatureTypeNotExistsError(f"Could not find feature type with name='{name}'!")
            return cast(FeatureTypeModel, FeatureTypeModel.from_orm(feature_type))



# flake8: noqa
from .client import StashHttpClient, StashHttpClientConflictError
from .models import (
    StashBranch,
    StashErrorResponse,
    StashPrCreationResponse,
    StashProject,
    StashPrRequest,
    StashRepository,
    StashReviewer,
    StashReviewerInfo,
)
from .settings import OverhaveStashClientSettings



from datetime import timedelta
from pathlib import Path

import typer

from overhave.base_settings import LoggingSettings
from overhave.transport import OverhaveS3Bucket, OverhaveS3ManagerSettings, S3Manager
from overhave.utils import get_current_time

s3_app = typer.Typer(short_help="Run s3 cloud interaction commands")

s3_bucket_app = typer.Typer(short_help="S3 cloud bucket's interaction commands")
s3_app.add_typer(s3_bucket_app, name="bucket")


def _check_bucket_registered(name: str) -> None:
    if name in (item.value for item in list(OverhaveS3Bucket)):
        return
    typer.secho(f"Note: specified s3 bucket name '{name}' not presented in OverhaveS3Bucket enum!", fg="yellow")


def _get_s3_manager() -> S3Manager:
    LoggingSettings().setup_logging()
    manager = S3Manager(OverhaveS3ManagerSettings(autocreate_buckets=False))
    manager.initialize()
    return manager


@s3_bucket_app.command(short_help="Create s3 cloud bucket")
def create(name: str = typer.Option(..., "-n", "--name", help="Declared s3 bucket")) -> None:
    """Create s3 bucket."""
    _check_bucket_registered(name)
    _get_s3_manager().create_bucket(name)


@s3_bucket_app.command(short_help="Delete s3 cloud bucket")
def delete(
    name: str = typer.Option(..., "-n", "--name", help="Declared s3 bucket"),
    force: bool = typer.Option(
        False, "-f", "--force", is_flag=True, help="Delete all files in bucket, then delete bucket"
    ),
) -> None:
    """Delete s3 bucket."""
    _check_bucket_registered(name)
    _get_s3_manager().delete_bucket(name, force=force)


@s3_bucket_app.command(short_help="Remove old s3 cloud bucket files")
def remove_files(
    name: str = typer.Option(..., "-n", "--name", help="Declared s3 bucket"),
    days: int = typer.Option(..., "-d", "--days", help="Remove all files in bucket older then specified days value"),
) -> None:
    """Remove s3 bucket files older than specified number of days."""
    _check_bucket_registered(name)
    manager = _get_s3_manager()
    target_date = get_current_time() - timedelta(days=days)

    objects = manager.get_bucket_objects(name)
    objects_to_delete = []
    for obj in objects:
        if not obj.modified_at < target_date:
            continue
        objects_to_delete.append(obj)
    if not objects_to_delete:
        typer.secho(f"No one object older than {days} days.")
        return
    typer.secho(f"Objects older then {days} days: {[x.name for x in objects_to_delete]}")
    manager.delete_bucket_objects(bucket=name, objects=objects_to_delete)


@s3_app.command(short_help="Download file from s3 bucket")
def download_file(
    bucket: str = typer.Option(..., "-b", "--bucket", help="Declared s3 bucket"),
    filename: str = typer.Option(..., "-f", "--filename", help="Filename for downloading"),
    dir_to_save: str = typer.Option(".", help="Directory for saving file"),
) -> None:
    """Create s3 bucket."""
    _check_bucket_registered(bucket)
    _get_s3_manager().download_file(filename=filename, bucket=bucket, dir_to_save=Path(dir_to_save))



# flake8: noqa
from .archiver import ArchiveManager
from .auth_managers import (
    DefaultAdminAuthorizationManager,
    IAdminAuthorizationManager,
    LDAPAdminAuthorizationManager,
    OverhaveLdapManagerSettings,
    SimpleAdminAuthorizationManager,
)
from .feature import FeatureExtractor, FeatureTypeExtractionError, IFeatureExtractor, ScenariosTestFileNotFound
from .file_extractor import BaseFileExtractor
from .git_initializer import GitPullError, GitRepositoryInitializationError, GitRepositoryInitializer
from .language import StepPrefixesModel
from .report_manager import ReportManager, ReportPresenceResolution
from .settings import (
    OverhaveAdminSettings,
    OverhaveDescriptionManagerSettings,
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveRedisSettings,
    OverhaveReportManagerSettings,
    OverhaveScenarioCompilerSettings,
    OverhaveStepContextSettings,
    ProcessorSettings,
)



from typing import List

import fastapi as fastapi

from overhave.api.auth import AuthToken, get_authorized_user
from overhave.api.views import (
    docs,
    favicon,
    get_features_handler,
    login_for_access_token,
    tags_item_handler,
    tags_list_handler,
    test_user_get_spec_handler,
    test_user_handler,
    test_user_put_spec_handler,
)
from overhave.storage import FeatureModel, TagModel, TestUserModel, TestUserSpecification


def _get_tags_router() -> fastapi.APIRouter:
    tags_router = fastapi.APIRouter()
    tags_router.add_api_route(
        "/item",
        tags_item_handler,
        methods=["GET"],
        response_model=TagModel,
        summary="Get FeatureTag",
        description="Get feature tag by `value`",
    )
    tags_router.add_api_route(
        "/list",
        tags_list_handler,
        response_model=List[TagModel],
        methods=["GET"],
        summary="Get FeatureTags",
        description="Get list of feature tags like `value`",
    )
    return tags_router


def _get_feature_router() -> fastapi.APIRouter:
    feature_router = fastapi.APIRouter()
    feature_router.add_api_route(
        "/",
        get_features_handler,
        methods=["GET"],
        response_model=List[FeatureModel],
        summary="Get list of Feature info",
        description="Get list of feature info by `tag_id` or `tag_value`",
    )
    return feature_router


def _get_testuser_router() -> fastapi.APIRouter:
    test_user_router = fastapi.APIRouter()
    test_user_router.add_api_route(
        "/",
        test_user_handler,
        methods=["GET"],
        response_model=TestUserModel,
        summary="Get TestUser",
        description="Get test user full info by `user_id` or `user_name`",
    )
    test_user_router.add_api_route(
        "/{user_id}/specification",
        test_user_get_spec_handler,
        methods=["GET"],
        response_model=TestUserSpecification,
        summary="Get TestUser specification",
        description="Get test user specification by `user_id`",
    )
    test_user_router.add_api_route(
        "/{user_id}/specification",
        test_user_put_spec_handler,
        methods=["PUT"],
        summary="Put test user specification",
        description="Update test user specification by `user_id` and given payload",
    )
    return test_user_router


def _get_auth_router() -> fastapi.APIRouter:
    auth_router = fastapi.APIRouter()
    auth_router.add_api_route(
        "/token",
        login_for_access_token,
        methods=["POST"],
        response_model=AuthToken,
        summary="Get OAuth2 token",
        description="Get OAuth2 token by username and password",
        include_in_schema=False,
    )
    return auth_router


def create_overhave_api() -> fastapi.FastAPI:
    app = fastapi.FastAPI()
    auth_deps = [fastapi.Depends(get_authorized_user)]

    app.include_router(_get_tags_router(), dependencies=auth_deps, prefix="/feature/tags", tags=["feature_tags"])
    app.include_router(_get_feature_router(), dependencies=auth_deps, prefix="/feature", tags=["features"])
    app.include_router(_get_testuser_router(), dependencies=auth_deps, prefix="/test_user", tags=["test_users"])

    app.include_router(_get_auth_router())
    app.add_api_route("/", docs, methods=["GET"], include_in_schema=False)
    app.add_api_route("/favicon.ico", favicon, methods=["GET"], include_in_schema=False)
    return app



# flake8: noqa
from .allure_utils import (
    DescriptionManager,
    StepContextNotDefinedError,
    StepContextRunner,
    add_scenario_title_to_report,
    get_description_manager,
    get_step_context_runner,
    set_severity_level,
)
from .basic import get_full_step_name, get_scenario, is_pytest_bdd_item
from .links import add_issue_links_to_report, has_issue_links, set_issue_links



from functools import cache

from overhave.factory.components import (
    AdminFactory,
    EmulationFactory,
    IAdminFactory,
    IEmulationFactory,
    IPublicationFactory,
    ISynchronizerFactory,
    ITestExecutionFactory,
    PublicationFactory,
    SynchronizerFactory,
    TestExecutionFactory,
)


@cache
def get_admin_factory() -> IAdminFactory:
    return AdminFactory()


@cache
def get_test_execution_factory() -> ITestExecutionFactory:
    return TestExecutionFactory()


@cache
def get_publication_factory() -> IPublicationFactory:
    return PublicationFactory()


@cache
def get_emulation_factory() -> IEmulationFactory:
    return EmulationFactory()


@cache
def get_synchronizer_factory() -> ISynchronizerFactory:
    return SynchronizerFactory()



import logging
from typing import Optional

from pydantic.types import SecretStr

from overhave.api import AuthToken
from overhave.transport.http.api_client.models import TokenRequestData
from overhave.transport.http.api_client.settings import OverhaveApiAuthenticatorSettings
from overhave.transport.http.api_client.storage import AuthStorage
from overhave.transport.http.base_client import BaseHttpClient, BearerAuth, HttpMethod

logger = logging.getLogger(__name__)


class OverhaveApiAuthenticator(BaseHttpClient[OverhaveApiAuthenticatorSettings]):
    """Client for auth_managers in Overhave API."""

    def __init__(
        self,
        settings: OverhaveApiAuthenticatorSettings,
        auth_storage: AuthStorage,
    ):
        super().__init__(settings=settings)
        self._auth_storage = auth_storage

    def _auth_by_credentials(self, username: str, password: SecretStr) -> AuthToken:
        logger.info("Get auth_managers token by username and password")
        data = TokenRequestData(username=username, password=password.get_secret_value())
        response = self._make_request(
            method=HttpMethod.POST,
            url=self._settings.get_auth_token_url,
            data=data.dict(by_alias=True),
        )
        return AuthToken.parse_obj(response.json())

    def _get_auth_token(self, username: str, password: SecretStr) -> AuthToken:
        token: Optional[AuthToken] = self._auth_storage.get_auth_token(username=username)
        if token is None:
            token = self._auth_by_credentials(username=username, password=password)
            self._auth_storage.update_auth_token(username=username, new_auth_token=token)
        return token

    def get_bearer_auth(self, username: str, password: SecretStr) -> BearerAuth:
        token = self._get_auth_token(username=username, password=password)
        return BearerAuth(token.access_token)



import typing
from pathlib import Path

from flask_admin.form import BaseForm
from flask_admin.model.widgets import XEditableWidget
from markupsafe import Markup
from wtforms import Field


class CustomPageWidget(XEditableWidget):
    """Customising widget.

    Widget content is filled with content from specified
    ``self.template_path```.
    """

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.template_path: typing.Optional[Path] = None
        super().__init__(*args, **kwargs)

    def __call__(self, field: Field, **kwargs: typing.Any) -> Markup:
        if self.template_path is not None:
            with self.template_path.open("r") as template:
                return Markup(template.read())
        return Markup(
            """
            <h1>Overhave index</h1>
            <p>This is Overhave info page. You could replace this page with your own HTML template.</p>
            <p>It is possible via <a href="https://overhave.readthedocs.io/en/latest/#context-setting" target="_blank">
            setting context</a> with <b>OverhaveAdminSettings</b>(index_template_path=<i>&lt;MY_TEMPLATE.html&gt;</i>)
            or through environment variable <b>OVERHAVE_INDEX_TEMPLATE_PATH</b>=<i>&lt;MY_TEMPLATE.html&gt;</i>.
            """
        )


class CustomPageField(Field):
    """Field for custom page widget."""

    widget = CustomPageWidget()


class CustomPageForm(BaseForm):
    """Form for custom page."""

    type = CustomPageField()

    def __init__(self, template_path: typing.Optional[Path]):
        super().__init__()
        self.type.widget.template_path = template_path
        self.type.label = None



import logging
from http import HTTPStatus
from typing import cast

from overhave.transport.http.base_client import (
    BaseHttpClient,
    BaseHttpClientException,
    BearerAuth,
    HttpClientValidationError,
    HttpMethod,
)
from overhave.transport.http.stash_client.models import STASH_RESPONSE_MODELS, AnyStashResponseModel, StashPrRequest
from overhave.transport.http.stash_client.settings import OverhaveStashClientSettings

logger = logging.getLogger(__name__)


class BaseStashHttpClientException(BaseHttpClientException):
    """Base exception for :class:`StashHttpClient`."""


class StashHttpClientConflictError(BaseStashHttpClientException):
    """Exception for situation with `HTTPStatus.CONFLICT` response.status_code."""


class StashHttpClient(BaseHttpClient[OverhaveStashClientSettings]):
    """Client for communication with remote Bitbucket server."""

    def send_pull_request(self, pull_request: StashPrRequest) -> AnyStashResponseModel:
        url = self._settings.get_pr_url(
            project_key=pull_request.target_branch.repository.project.key,
            repository_name=pull_request.target_branch.repository.name,
        )
        response = self._make_request(
            method=HttpMethod.POST,
            url=url,
            json=pull_request.dict(by_alias=True),
            auth=BearerAuth(self._settings.auth_token),
            raise_for_status=False,
        )
        if response.status_code == HTTPStatus.CONFLICT:
            raise StashHttpClientConflictError("Got conflict when trying to send pull-request!")
        response.raise_for_status()
        for model in STASH_RESPONSE_MODELS:
            try:
                logger.debug("Trying to parse '%s'...", model)
                return cast(AnyStashResponseModel, self._parse_or_raise(response, model))
            except HttpClientValidationError:
                logger.debug("Could not convert response to '%s'!", model, exc_info=True)
        raise HttpClientValidationError(
            f"Could not parse Stash response while trying to create pull-request!\nResponse: {response.json()}"
        )



from typing import List, Optional, Tuple, cast

from pydantic.main import BaseModel
from pytest_bdd import types as default_types
from pytest_bdd.parser import STEP_PREFIXES


class StepPrefixesModel(BaseModel):
    """Base model for default STEP_PREFIXES extension."""

    FEATURE: str
    SCENARIO_OUTLINE: str
    SCENARIO: str
    BACKGROUND: str
    EXAMPLES: str
    EXAMPLES_VERTICAL: str
    GIVEN: str
    WHEN: str
    THEN: str
    AND: str
    BUT: str

    def extend_defaults(self) -> List[Tuple[str, Optional[str]]]:
        """Extend default STEP_PREFIXES from pytest_bdd."""
        STEP_PREFIXES.append((self.FEATURE, default_types.FEATURE))
        STEP_PREFIXES.append((self.SCENARIO_OUTLINE, default_types.SCENARIO_OUTLINE))
        STEP_PREFIXES.append((self.SCENARIO, default_types.SCENARIO))
        STEP_PREFIXES.append((self.BACKGROUND, default_types.BACKGROUND))

        STEP_PREFIXES.append((self.EXAMPLES, default_types.EXAMPLES))
        STEP_PREFIXES.append((self.EXAMPLES_VERTICAL, default_types.EXAMPLES_VERTICAL))

        STEP_PREFIXES.append((self.GIVEN, default_types.GIVEN))
        STEP_PREFIXES.append((self.WHEN, default_types.WHEN))
        STEP_PREFIXES.append((self.THEN, default_types.THEN))

        STEP_PREFIXES.append((self.AND, None))
        STEP_PREFIXES.append((self.BUT, None))

        return cast(List[Tuple[str, Optional[str]]], STEP_PREFIXES)



from datetime import datetime

from pydantic.fields import Field
from pydantic.main import BaseModel


class AuthToken(BaseModel):
    """Model for OAuth2 auth_managers token."""

    access_token: str
    expires_at: datetime
    token_type: str = Field("Bearer", const=True)


class AuthTokenData(BaseModel):
    """Model for OAuth2 auth_managers token data."""

    username: str


class AuthHeaders(BaseModel):
    """Model for OAuth2 auth_managers HTTP headers."""

    Authorization: str = Field("Bearer", const=True)


AUTH_HEADERS = AuthHeaders()



from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

from flask_admin.contrib.sqla import ModelView
from pydantic import root_validator, validator
from pydantic.datetime_parse import timedelta
from yarl import URL

from overhave.base_settings import BaseOverhavePrefix
from overhave.entities.language import StepPrefixesModel


class OverhaveAdminSettings(BaseOverhavePrefix):
    """Settings for Overhave Flask Admin customization."""

    # Path to custom index template. By default, contains Overhave project info.
    index_template_path: Optional[Path]

    # Custom SQLAlchemy ModelViews for Overhave admin panel
    custom_views: Optional[Sequence[ModelView]]

    # Enable testing with test execution consumer, based on Redis tasks. Enabled by default.
    # When disabled - all test runs will be executed with :class:`Threadpool`.
    consumer_based: bool = True

    # Threadpool size for admin service
    threadpool_process_num: int = 5


class OverhaveScenarioCompilerSettings(BaseOverhavePrefix):
    """Settings for scenario compiling and parsing."""

    tag_prefix: str = "@"
    id_prefix: str = "ID:"
    created_by_prefix: str = "# created by"
    last_edited_by_prefix: str = "last edited by"
    published_by_prefix: str = "published by"
    severity_prefix: str = "@severity."

    time_delimiter: str = ","
    time_format: str = "%d-%m-%Y %H:%M:%S"
    blocks_delimiter: str = "|"

    @property
    def severity_keyword(self) -> str:
        return self.severity_prefix.removeprefix(self.tag_prefix)


class OverhaveLanguageSettings(BaseOverhavePrefix):
    """Settings for language definitions."""

    step_prefixes: Optional[StepPrefixesModel]


class OverhaveFileSettings(BaseOverhavePrefix):
    """Settings for scenario file savings."""

    feature_suffix: str = ".feature"
    fixture_suffix: str = ".py"

    # Current workdir, used for :class:`PluginResolver` for creating of relative pytest plugins' paths
    work_dir: Path = Path.cwd()

    # Root project directory with features, fixtures and steps packages
    root_dir: Optional[Path]

    # Base directory for feature files, by default - root_dir / 'features'
    features_dir: Path

    # Base directory for pytest files with template mask, by default - root_dir / 'fixtures'
    fixtures_dir: Path
    # Template mask for fixtures pytest files which contain `feature_type` key
    fixtures_file_template_mask: str = "test_{feature_type}.py"

    # Flag for `steps_dir` validation in case of relating to `work_dir`
    validate_steps_dir: bool = False

    # Base directory for pytest-bdd steps, , by default - root_dir / 'steps'
    steps_dir: Path

    # Temporary directory for scenarios test runs
    tmp_dir: Path = Path("/tmp/overhave")  # noqa: S108

    @root_validator(pre=True)
    def validate_dirs(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        root_dir = values.get("root_dir")
        if root_dir:
            for directory in ("features_dir", "fixtures_dir", "steps_dir"):
                if values.get(directory):
                    continue
                values[directory] = Path(root_dir) / directory.replace("_dir", "")
        return values

    @validator("steps_dir")
    def validate_nesting(cls, v: Path, values: Dict[str, Any]) -> Path:
        validate_steps_dir = values["validate_steps_dir"]
        if validate_steps_dir:
            work_dir: Path = values["work_dir"]
            v.relative_to(work_dir)
        return v

    @property
    def tmp_features_dir(self) -> Path:
        return self.tmp_dir / "features"

    @property
    def tmp_fixtures_dir(self) -> Path:
        return self.tmp_dir / "fixtures"

    @property
    def tmp_reports_dir(self) -> Path:
        return self.tmp_dir / "reports"


class OverhaveReportManagerSettings(BaseOverhavePrefix):
    """Settings for :class:`ReportManager`."""

    report_creation_timeout: int = 120  # sec
    report_creation_error_msg: str = "not_created"
    allure_cmdline: str = "allure"
    archive_extension: str = "zip"


class ProcessorSettings(BaseOverhavePrefix):
    """Settings for :class:`Processor`."""

    processes_num: int = 5


class OverhaveRedisSettings(BaseOverhavePrefix):
    """Settings for Redis entities, which use for work with different framework tasks."""

    redis_url: URL = URL("redis://localhost:6379")
    redis_db: int = 0
    redis_block_timeout: timedelta = timedelta(seconds=1)
    redis_read_count: int = 1

    @validator("redis_url", pre=True)
    def validate_url(cls, v: Union[str, URL]) -> URL:
        if isinstance(v, str):
            return URL(v)
        return v

    @property
    def timeout_milliseconds(self) -> int:
        return int(self.redis_block_timeout.total_seconds() * 1000)


class OverhaveEmulationSettings(BaseOverhavePrefix):
    """Settings for Overhave Emulator, which emulates session with test user."""

    # Path for emulation core application, such as GoTTY
    emulation_core_path: str = "/gotty"

    # All emulation core application prefixes for execution
    emulation_prefix: str = "--permit-write --once --address {address} --port {port} --timeout {timeout}"

    # Specific terminal tool startup command with relative `feature_type`, for example: `myapp {feature_type}`
    emulation_base_cmd: Optional[str]
    # Terminal tool command postfix with specified user `name` and `model`, for example: `--name={name} --model={model}`
    # If it is no need in use - may be optional.
    emulation_postfix: Optional[str]

    # Optional additional terminal tool usage description
    emulation_desc_link: Optional[str]

    emulation_bind_ip: str = "0.0.0.0"  # noqa: S104
    # Ports for emulation binding. Expects as string with format `["port1", "port2", ...]`
    emulation_ports: List[int] = [8080]

    # As a real service, should be used follow path: `http://my-service.domain/mount`
    # where `emulation_service_url` = `http://my-service.domain` - URL for service,
    # and `emulation_service_mount` = `mount` - mount point for service redirection.
    # If `emulation_service_mount` is `None` - this is localhost debug.
    emulation_service_url: URL = URL("http://localhost")
    emulation_service_mount: Optional[str]

    # Wait until emulation become served
    emulation_wait_timeout: timedelta = timedelta(seconds=300)

    @validator("emulation_service_url", pre=True)
    def validate_url(cls, v: Union[str, URL]) -> URL:
        if isinstance(v, str):
            return URL(v)
        return v

    @validator("emulation_ports", pre=True)
    def validate_ports(cls, v: Union[List[int], str]) -> List[int]:
        if isinstance(v, str):
            return [int(x.strip()) for x in v.split(",")]
        return v

    def get_emulation_url(self, port: str) -> str:
        if isinstance(self.emulation_service_mount, str):
            return (
                self.emulation_service_url / self.emulation_service_mount / port / ""
            ).human_repr()  # NGINX should redirect to `service:port` by mount point and specified port
        return self.emulation_service_url.with_port(int(port)).human_repr()  # only for local debug

    @property
    def wait_timeout_seconds(self) -> int:
        return int(self.emulation_wait_timeout.total_seconds())

    @property
    def enabled(self) -> bool:
        return isinstance(self.emulation_base_cmd, str)


class OverhaveStepContextSettings(BaseOverhavePrefix):
    """Settings for :class:`StepContextRunner`."""

    step_context_logs: bool = False


class OverhaveDescriptionManagerSettings(BaseOverhavePrefix):
    """Settings for DescriptionManager, which sets the description to Allure report dynamically after test."""

    blocks_delimiter: str = ""
    html: bool = True



import abc
import logging

from overhave.entities.auth_managers.abstract import IAdminAuthorizationManager
from overhave.storage import ISystemUserStorage

logger = logging.getLogger(__name__)


class BaseAdminAuthorizationManager(IAdminAuthorizationManager, abc.ABC):
    """Base class for user auth_managers."""

    def __init__(self, system_user_storage: ISystemUserStorage):
        self._system_user_storage = system_user_storage
        logger.info("Authorization manager: %s", type(self))



# flake8: noqa
from .api_client import AuthStorage, IAuthStorage, OverhaveApiAuthenticator, OverhaveApiAuthenticatorSettings
from .base_client import BaseHttpClient, BaseHttpClientSettings, BearerAuth
from .gitlab_client import (
    GitlabHttpClient,
    GitlabHttpClientConflictError,
    GitlabMrCreationResponse,
    GitlabMrRequest,
    GitlabRepository,
    OverhaveGitlabClientSettings,
)
from .stash_client import (
    OverhaveStashClientSettings,
    StashBranch,
    StashErrorResponse,
    StashHttpClient,
    StashHttpClientConflictError,
    StashPrCreationResponse,
    StashProject,
    StashPrRequest,
    StashRepository,
    StashReviewer,
    StashReviewerInfo,
)



import logging

import git

from overhave.entities.settings import OverhaveFileSettings

logger = logging.getLogger(__name__)


class BaseGitRepositoryInitializerException(Exception):
    """Base exception for GitRepositoryInitializer."""


class GitRepositoryInitializationError(BaseGitRepositoryInitializerException):
    """Error for situation with `git.GitError` when repo initializing."""


class GitPullError(BaseGitRepositoryInitializerException):
    """Error for situation with `git.GitError` when origin pulling."""


class GitRepositoryInitializer:
    """Class for git repository initializing and updating with pull command."""

    def __init__(self, file_settings: OverhaveFileSettings) -> None:
        self._file_settings = file_settings

        self._init_repository()

    def _init_repository(self) -> None:
        try:
            logger.debug("Initialize git repository in '%s'...", self._file_settings.features_dir)
            self._repository = git.Repo(self._file_settings.features_dir)
            logger.info("Repository: %s", self._repository)
            logger.info("Working dir: %s", self._repository.working_dir)
            logger.info("Active head: %s", self._repository.active_branch)
        except git.GitError as err:
            raise GitRepositoryInitializationError("Error while trying to initialize git repository!") from err

    @property
    def repository(self) -> git.Repo:
        return self._repository

    def pull(self) -> None:
        try:
            origin = self._repository.remotes.origin
            origin.pull()
            logger.info("Origin successfully pulled")
        except git.GitError as err:
            raise GitPullError("Error while trying to pull from remote repository!") from err



from http import HTTPStatus
from pathlib import Path

import fastapi
import starlette.responses

from overhave.api.deps import get_admin_files_dir


async def favicon(files_dir: Path = fastapi.Depends(get_admin_files_dir)) -> starlette.responses.FileResponse:
    icon_path = files_dir / "favicon.ico"
    if icon_path.exists():
        return starlette.responses.FileResponse(icon_path.as_posix())
    raise fastapi.HTTPException(status_code=HTTPStatus.NOT_FOUND)


async def docs() -> starlette.responses.RedirectResponse:
    return starlette.responses.RedirectResponse(url="/docs")



from functools import cache
from pathlib import Path

from overhave.api.settings import OverhaveApiAuthSettings
from overhave.storage import (
    DraftStorage,
    FeatureStorage,
    FeatureTagStorage,
    IDraftStorage,
    IFeatureStorage,
    IFeatureTagStorage,
    ISystemUserStorage,
    ITestUserStorage,
    SystemUserStorage,
    TestUserStorage,
)


@cache
def get_api_auth_settings() -> OverhaveApiAuthSettings:
    return OverhaveApiAuthSettings()


@cache
def get_admin_files_dir() -> Path:
    return Path(__file__).parent.parent / "admin" / "files"


@cache
def get_system_user_storage() -> ISystemUserStorage:
    return SystemUserStorage()


@cache
def get_feature_tag_storage() -> IFeatureTagStorage:
    return FeatureTagStorage()


@cache
def get_feature_storage() -> IFeatureStorage:
    return FeatureStorage()


@cache
def get_draft_storage() -> IDraftStorage:
    return DraftStorage()


@cache
def get_test_user_storage() -> ITestUserStorage:
    return TestUserStorage()



# flake8: noqa
from overhave.admin import OverhaveAdminApp, overhave_app
from overhave.api import create_overhave_api as overhave_api
from overhave.base_settings import AuthorizationStrategy as OverhaveAuthorizationStrategy
from overhave.base_settings import DataBaseSettings as OverhaveDBSettings
from overhave.base_settings import LoggingSettings as OverhaveLoggingSettings
from overhave.base_settings import OverhaveAuthorizationSettings
from overhave.cli import group, set_config_to_context
from overhave.entities import (
    OverhaveAdminSettings,
    OverhaveDescriptionManagerSettings,
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveLdapManagerSettings,
    OverhaveRedisSettings,
    OverhaveScenarioCompilerSettings,
    OverhaveStepContextSettings,
    StepPrefixesModel,
)
from overhave.factory import ConsumerFactory as OverhaveConsumerFactory
from overhave.factory import IAdminFactory as OverhaveAdminFactory
from overhave.factory import IEmulationFactory as OverhaveEmulationFactory
from overhave.factory import IPublicationFactory as OverhavePublicationFactory
from overhave.factory import ISynchronizerFactory as OverhaveSynchronizerFactory
from overhave.factory import ITestExecutionFactory as OverhaveTestExecutionFactory
from overhave.factory import (
    OverhaveAdminContext,
    OverhaveEmulationContext,
    OverhavePublicationContext,
    OverhaveSynchronizerContext,
    OverhaveTestExecutionContext,
)
from overhave.factory import get_admin_factory as overhave_admin_factory
from overhave.factory import get_emulation_factory as overhave_emulation_factory
from overhave.factory import get_publication_factory as overhave_publication_factory
from overhave.factory import get_synchronizer_factory as overhave_synchronizer_factory
from overhave.factory import get_test_execution_factory as overhave_test_execution_factory
from overhave.publication import OverhaveGitlabPublisherSettings, OverhaveStashPublisherSettings
from overhave.publication import PublicationManagerType as OverhavePublicationManagerType
from overhave.publication import PublicationSettings as OverhavePublicationSettings
from overhave.pytest_plugin import IProxyManager as OverhaveProxyManager
from overhave.pytest_plugin import get_description_manager
from overhave.pytest_plugin import get_proxy_manager as overhave_proxy_manager
from overhave.storage import FeatureTypeName as OverhaveFeatureName
from overhave.storage import TestUserSpecification, TestUserStorage
from overhave.test_execution import OverhaveProjectSettings, OverhaveTestSettings
from overhave.transport import (
    OverhaveApiAuthenticator,
    OverhaveApiAuthenticatorSettings,
    OverhaveGitlabClientSettings,
    OverhaveLdapClientSettings,
    OverhaveS3ManagerSettings,
    OverhaveStashClientSettings,
)
from overhave.transport import RedisStream as OverhaveRedisStream



import logging
from typing import cast

from gitlab.exceptions import GitlabAuthenticationError

from overhave.transport.http import BaseHttpClient
from overhave.transport.http.base_client import BaseHttpClientException
from overhave.transport.http.gitlab_client.models import GitlabMrCreationResponse, GitlabMrRequest
from overhave.transport.http.gitlab_client.settings import OverhaveGitlabClientSettings
from overhave.transport.http.gitlab_client.utils import get_gitlab_python_client

logger = logging.getLogger(__name__)


class BaseGitlabHttpClientException(BaseHttpClientException):
    """Base exception for :class:`GitlabHttpClient`."""


class GitlabHttpClientConflictError(BaseGitlabHttpClientException):
    """Exception for situation with `HTTPStatus.CONFLICT` response.status_code."""


class GitlabInvalidTokenError(BaseGitlabHttpClientException):
    """Exception for situation with invalid token or gitlab url."""


class GitlabHttpClient(BaseHttpClient[OverhaveGitlabClientSettings]):
    """Client for communication with remote Gitlab server."""

    def send_merge_request(
        self, repository_id: str, merge_request: GitlabMrRequest, token: str
    ) -> GitlabMrCreationResponse:
        gitlab_python_client = get_gitlab_python_client(
            url=self._settings.url.human_repr(),
            token_type=self._settings.token_type,
            token=token,
        )
        project = gitlab_python_client.projects.get(repository_id, lazy=True)
        try:
            return cast(GitlabMrCreationResponse, project.mergerequests.create(merge_request.dict(by_alias=True)))
        except GitlabAuthenticationError as e:
            logging.exception("Please verify your token or URL! Maybe they are invalid")
            raise GitlabInvalidTokenError("Please verify your token or URL! Maybe they are invalid") from e



import typer

from overhave.base_settings import DataBaseSettings, LoggingSettings
from overhave.cli.group import overhave
from overhave.factory import get_synchronizer_factory
from overhave.synchronization import IOverhaveSynchronizer


def _create_synchronizer() -> IOverhaveSynchronizer:
    DataBaseSettings().setup_db()
    LoggingSettings().setup_logging()
    return get_synchronizer_factory().synchronizer


@overhave.command(short_help="Run Overhave feature synchronization")
def synchronize(
    create_db_features: bool = typer.Option(
        False, "-c", "--create-db-features", is_flag=True, help="Create features in database if necessary"
    ),
    pull_repository: bool = typer.Option(False, "-p", "--pull-repository", is_flag=True, help="Pull remote repository"),
) -> None:
    _create_synchronizer().synchronize(create_db_features, pull_repository)



import logging
from datetime import datetime
from pathlib import Path
from typing import List

import allure
import pytz

from overhave.entities import BaseFileExtractor, FeatureExtractor, GitRepositoryInitializer, OverhaveFileSettings
from overhave.scenario import ScenarioParser
from overhave.scenario.parser import FeatureInfo
from overhave.storage import (
    FeatureModel,
    IDraftStorage,
    IFeatureStorage,
    IFeatureTagStorage,
    IFeatureTypeStorage,
    IScenarioStorage,
    ISystemUserStorage,
    ScenarioModel,
    TagModel,
)
from overhave.synchronization.abstract import IOverhaveSynchronizer

logger = logging.getLogger(__name__)


class BaseOverhaveSynchronizerException(Exception):
    """Base exception for :class:`OverhaveSynchronizer`."""


class NullableLastEditedAtError(BaseOverhaveSynchronizerException):
    """Exception for nullable last_edited_at."""


class NullableInfoLastEditedByError(BaseOverhaveSynchronizerException):
    """Exception for nullable feature info last_edited_at."""


class NullableInfoNameError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info name."""


class NullableInfoScenariosError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info scenarios."""


class NullableInfoAuthorError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info author."""


class NullableInfoFeatureTypeError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info type."""


class FeatureInfoUserNotFoundError(BaseOverhaveSynchronizerException):
    """Exception for situation without specified user in database."""


class OverhaveSynchronizer(BaseFileExtractor, IOverhaveSynchronizer):
    """Class for synchronization between git and database."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        feature_storage: IFeatureStorage,
        feature_type_storage: IFeatureTypeStorage,
        scenario_storage: IScenarioStorage,
        draft_storage: IDraftStorage,
        tag_storage: IFeatureTagStorage,
        scenario_parser: ScenarioParser,
        feature_extractor: FeatureExtractor,
        system_user_storage: ISystemUserStorage,
        git_initializer: GitRepositoryInitializer,
    ):
        super().__init__(extenstion=file_settings.feature_suffix)
        self._file_settings = file_settings
        self._feature_storage = feature_storage
        self._feature_type_storage = feature_type_storage
        self._scenario_storage = scenario_storage
        self._draft_storage = draft_storage
        self._tag_storage = tag_storage
        self._scenario_parser = scenario_parser
        self._feature_extractor = feature_extractor
        self._system_user_storage = system_user_storage
        self._git_initializer = git_initializer

    @staticmethod
    def _update_feature_model_with_info(
        model: FeatureModel, info: FeatureInfo, file_ts: datetime, tags: List[TagModel]
    ) -> None:
        if info.name is None:
            raise NullableInfoNameError("Feature info has not got feature name!")
        model.name = info.name
        if info.last_edited_by is not None:
            model.last_edited_by = info.last_edited_by
        model.last_edited_at = file_ts
        if info.tasks is not None:
            model.task = info.tasks
        model.feature_tags = tags

    def _get_last_change_time(self, model: FeatureModel) -> datetime:
        draft_models = self._draft_storage.get_drafts_by_feature_id(model.id)
        if draft_models:
            logger.info("Feature has got drafts.")
            last_published_draft = draft_models[-1]
            if last_published_draft.published_at is not None:
                logger.info(
                    "Last version has been published at %s.",
                    last_published_draft.published_at.strftime("%d-%m-%Y %H:%M:%S"),
                )
                return last_published_draft.published_at
        logger.info("Feature hasn't got any published version.")
        return model.last_edited_at

    def _get_feature_tags(self, info: FeatureInfo) -> List[TagModel]:
        tags: List[TagModel] = []
        if info.tags is not None:
            if info.last_edited_by is None:
                raise NullableInfoLastEditedByError("last_edited_by value should not be None!")
            for tag in info.tags:
                tag_model = self._tag_storage.get_or_create_tag(value=tag, created_by=info.last_edited_by)
                tags.append(tag_model)
        return tags

    def _ensure_users_exist(self, info: FeatureInfo) -> None:
        for user in (x for x in {info.author, info.last_edited_by} if x):
            if self._system_user_storage.get_user_by_credits(login=user) is not None:
                continue
            raise FeatureInfoUserNotFoundError(f"Could not find user with login={user}!")

    def _update_db_feature(self, model: FeatureModel, info: FeatureInfo, file_ts: datetime) -> None:
        logger.info("Feature is gonna be updated...")
        if info.scenarios is None:
            raise NullableInfoScenariosError("Some troubles with parsing feature - could not get scenarios!")
        self._ensure_users_exist(info)
        feature_tags = self._get_feature_tags(info=info)
        self._update_feature_model_with_info(model=model, info=info, file_ts=file_ts, tags=feature_tags)
        scenario_model = self._scenario_storage.get_scenario_by_feature_id(model.id)
        self._feature_storage.update_feature(model)
        scenario_model.text = info.scenarios
        self._scenario_storage.update_scenario(model=scenario_model)
        logger.info("Feature has been updated successfully.")

    def _create_feature(self, file: Path, info: FeatureInfo) -> None:
        logger.info("Feature is gonna be created...")
        if info.name is None:
            raise NullableInfoNameError("Feature info has not got feature name!")
        if info.type is None:
            raise NullableInfoFeatureTypeError("Could not create feature without feature type!")
        if info.author is None:
            raise NullableInfoAuthorError("Could not create feature without author!")
        self._ensure_users_exist(info)
        feature_tags = self._get_feature_tags(info=info)
        feature_type = self._feature_type_storage.get_feature_type_by_name(info.type)
        if info.last_edited_by is None:
            info.last_edited_by = info.author
        feature_model = FeatureModel(
            id=0,
            name=info.name,
            author=info.author,
            type_id=feature_type.id,
            last_edited_by=info.last_edited_by,
            last_edited_at=info.last_edited_at,
            task=info.tasks or [],
            file_path=file.relative_to(
                self._feature_extractor.feature_type_to_dir_mapping[feature_type.name]
            ).as_posix(),
            released=True,
            feature_type=feature_type,
            feature_tags=feature_tags,
            severity=info.severity or allure.severity_level.NORMAL,
        )
        feature_model.id = self._feature_storage.create_feature(feature_model)
        scenario_model = ScenarioModel(id=0, feature_id=feature_model.id, text=info.scenarios)
        self._scenario_storage.create_scenario(scenario_model)
        logger.info("Feature with ID=%s has been created successfully.", feature_model.id)

    def synchronize(self, create_db_features: bool = False, pull_repository: bool = False) -> None:  # noqa: C901
        if pull_repository:
            self._git_initializer.pull()
        logger.info("Start synchronization...")
        all_features = self._extract_recursively(self._file_settings.features_dir)
        for feature_file in all_features:
            logger.info("Synchronize feature from file %s...", feature_file.as_posix())
            feature_info = self._scenario_parser.parse(feature_file.read_text())
            logger.debug("Parsed feature info: %s", feature_info)
            if feature_info.id is None:
                if not create_db_features:
                    logger.warning("Feature doesn't have Overhave ID or ID format is incorrect. Skip.")
                    continue
                self._create_feature(file=feature_file, info=feature_info)
                continue
            logger.info("Feature has ID=%s", feature_info.id)
            feature_model = self._feature_storage.get_feature(feature_info.id)
            if feature_model is None:
                logger.warning("Feature doesn't exist in Overhave database.")
                continue  # TODO: unlink file and create MR with deletions at the end

            feature_file_ts = datetime.fromtimestamp(feature_file.stat().st_mtime, tz=pytz.UTC)
            if feature_model.last_edited_at is None:
                raise NullableLastEditedAtError("last_edited_at value should not be None!")
            if feature_model.last_edited_at == feature_file_ts and feature_model.released:
                logger.warning("Feature has been already synchronized.")
                continue

            last_change_time = self._get_last_change_time(model=feature_model)
            if last_change_time < feature_file_ts:
                self._update_db_feature(model=feature_model, info=feature_info, file_ts=feature_file_ts)
                continue
            if feature_model.released:
                logger.info("Feature is already actual. Skip.")
                continue
            logger.info(
                "Feature was changed soon (at %s), but not released. Skip.",
                last_change_time.strftime("%d-%m-%Y %H:%M:%S"),
            )
        logger.info("Synchronization completed.")



from flask_admin import Admin
from flask_admin.menu import MenuLink

from overhave import db
from overhave.admin.views import (
    DraftView,
    EmulationRunView,
    EmulationView,
    FeatureView,
    GroupView,
    OverhaveIndexView,
    TagsView,
    TestRunView,
    TestUserView,
    UserView,
)
from overhave.factory import IAdminFactory


def get_flask_admin(factory: IAdminFactory) -> Admin:
    index_view = OverhaveIndexView(
        name="Info",
        url="/",
        auth_manager=factory.auth_manager,
        index_template_path=factory.context.admin_settings.index_template_path,
    )
    admin = Admin(name="Overhave", template_mode="bootstrap3", index_view=index_view)
    admin.add_link(MenuLink(name="Log out", url="/logout"))
    views_list = [
        FeatureView(db.Feature, db.current_session, category="Scenarios", name="Features"),
        TestRunView(db.TestRun, db.current_session, category="Scenarios", name="Test runs"),
        DraftView(db.Draft, db.current_session, category="Scenarios", name="Versions"),
        TagsView(db.Tags, db.current_session, category="Scenarios", name="Tags"),
        TestUserView(db.TestUser, db.current_session, name="Test users"),
        EmulationView(db.Emulation, db.current_session, category="Emulation", name="Templates"),
        EmulationRunView(db.EmulationRun, db.current_session, category="Emulation", name="Emulation runs"),
        UserView(db.UserRole, db.current_session, category="Access", name="Users"),
        GroupView(db.GroupRole, db.current_session, category="Access", name="Groups"),
    ]
    if factory.context.admin_settings.custom_views is not None:
        views_list.extend(factory.context.admin_settings.custom_views)
    admin.add_views(*views_list)
    return admin



import re
from functools import cache
from pathlib import Path
from typing import Any, Optional, Pattern

import allure
from _pytest.nodes import Item

from overhave.pytest_plugin.helpers.basic import get_scenario


def _get_severity_level_from_tags(item: Item, keyword: str) -> Optional[allure.severity_level]:
    for marker in reversed(item.own_markers):
        if not marker.name.startswith(keyword):
            continue
        severity = marker.name.removeprefix(keyword)
        return allure.severity_level(severity)
    return None


@cache
def _get_severity_pattern(keyword: str) -> Pattern[str]:
    return re.compile(rf"({keyword})(?P<severity>\w+)\b")


def _get_severity_level_from_feature(item: Item, keyword: str) -> Optional[allure.severity_level]:
    scenario = get_scenario(item)
    with Path(scenario.feature.filename).open() as feature_file:
        head_line = next(iter(feature_file))
        parsed_severity = _get_severity_pattern(keyword).search(head_line)
        if parsed_severity is not None:
            return allure.severity_level(parsed_severity.group("severity"))
    return None


def _get_default_severity(*args: Any, **kwargs: Any) -> allure.severity_level:
    return allure.severity_level.NORMAL


def set_severity_level(item: Item, keyword: str) -> None:
    for extractor_func in _get_severity_level_from_tags, _get_severity_level_from_feature, _get_default_severity:
        parsed_severity = extractor_func(*(item, keyword))  # type: ignore
        if parsed_severity is None:
            continue
        allure.dynamic.severity(parsed_severity)
        return



from typing import Any

import requests


class BearerAuth(requests.auth.AuthBase):
    """Class for Bearer auth_managers pattern."""

    def __init__(self, token: str) -> None:
        super().__init__()
        self.token = token

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BearerAuth):
            return False
        return self.token == other.token



class FeatureTypeExtractionError(RuntimeError):
    """Exception for feature type extraction error."""


class ScenariosTestFileNotFound(FileNotFoundError):
    """Exception for situation without scenarios test files."""



# flake8: noqa
from .abstract_publisher import IVersionPublisher
from .gitlab import GitlabVersionPublisher, OverhaveGitlabPublisherSettings, TokenizerClient, TokenizerClientSettings
from .objects import PublicationManagerType
from .settings import PublicationSettings
from .stash import OverhaveStashPublisherSettings, StashVersionPublisher
